/*************************************************************************
 *
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 * 
 * Copyright 2000, 2010 Oracle and/or its affiliates.
 *
 * OpenOffice.org - a multi-platform office productivity suite
 *
 * This file is part of OpenOffice.org.
 *
 * OpenOffice.org is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 3
 * only, as published by the Free Software Foundation.
 *
 * OpenOffice.org is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3 for more details
 * (a copy is included in the LICENSE file that accompanied this code).
 *
 * You should have received a copy of the GNU Lesser General Public License
 * version 3 along with OpenOffice.org.  If not, see
 * <http://www.openoffice.org/license.html>
 * for a copy of the LGPLv3 License.
 *
 ************************************************************************/

/*
 * TrueTypeCreator method implementation
 *
 * @author: Alexander Gelfenbain
 *
 */

#if OSL_DEBUG_LEVEL == 0
#  ifndef NDEBUG
#    define NDEBUG
#  endif
#endif
#include <assert.h>

#include "ttcr.hxx"
#include "string.h"


namespace vcl
{

//----------------------------------------------------------------------

TrueTypeCreator::TrueTypeCreator()
    : mpFontDirectory( NULL )
{
}

TrueTypeCreator::~TrueTypeCreator()
{
    TTT_TableMap::iterator it;
    for (it = maTableMap.begin(); it != maTableMap.end(); it++)
        delete it->second;

    delete mpFontDirectory;
}

void TrueTypeCreator::AddTable( TrueTypeTable* pTable )
{
    sal_uInt32 nTag = pTable->GetTag();
    maTableMap[ nTag ] = pTable;
}

void TrueTypeCreator::RemoveTable( sal_uInt32 nTag )
{
    maTableMap.erase( nTag );
}

TrueTypeTable* TrueTypeCreator::FindTable( sal_uInt32 nTag )
{
    TTT_TableMap::iterator it = maTableMap.find( nTag );
    return ( (it != maTableMap.end()) ? it->second : NULL );
}

sal_uInt32 TrueTypeCreator::CalculateFileSize()
{
    sal_uInt16 numTables = (sal_uInt16) maTableMap.size();
    sal_uInt32 s = 12 + 16 * numTables;

    TTT_TableMap::iterator it;
    for (it = maTableMap.begin(); it != maTableMap.end(); it++) {
        s += it->second->GetContainerSize();
    }

    return s;
}

/* This function processes all the tables and synchronizes them before creating
 * the output TrueType stream.
 *
 * *** It adds two TrueType tables to the font: 'loca' and 'hmtx' ***
 *
 * It does:
 *
 * - Re-numbers glyph IDs and creates 'glyf', 'loca', and 'hmtx' tables.
 * - Calculates xMin, yMin, xMax, and yMax and stores values in 'head' table.
 * - Stores indexToLocFormat in 'head'
 * - updates 'maxp' table
 * - Calculates advanceWidthMax, minLSB, minRSB, xMaxExtent and numberOfHMetrics
 *   in 'hhea' table
 *
 */

void TrueTypeCreator::ProcessTables()
{
    TrueTypeTable *maxp, *hhea, *loca, *hmtx;
    TTT_glyf* glyf;
    TTT_head* head;
    sal_uInt32 nGlyphs, locaLen = 0, glyfLen = 0;
    sal_Int16 xMin = 0, yMin = 0, xMax = 0, yMax = 0;
    sal_uInt32 i = 0;
    sal_Int16 indexToLocFormat;
    sal_uInt8 *hmtxPtr, *hheaPtr;
    sal_uInt32 hmtxSize;
    sal_uInt8 *p1, *p2;
    sal_uInt16 maxPoints = 0, maxContours = 0, maxCompositePoints = 0, maxCompositeContours = 0;
    int nlsb = 0;
    sal_uInt32 *gid;                        /* array of old glyphIDs */

    glyf = dynamic_cast< TTT_glyf* >( FindTable( T_glyf ) );
    const GlyphDataPtrMap& m = glyf->GetGlyphDataPtrMap();
    nGlyphs = m.size();
    assert(nGlyphs != 0);
    gid = (sal_uInt32*)scalloc(nGlyphs, sizeof(sal_uInt32));

    RemoveTable( T_loca );
    RemoveTable( T_hmtx );

    /* XXX Need to make sure that composite glyphs do not break during glyph renumbering */

    GlyphDataPtrMap::const_iterator it;
    for (it = m.begin(); it != m.end(); it++) {
        GlyphData *gd = it->second;
        sal_Int16 z;
        glyfLen += gd->nbytes;
        /* XXX if (gd->nbytes & 1) glyfLen++; */


        assert(gd->newID == i);
        gid[i++] = gd->glyphID;
        /* gd->glyphID = i++; */

        /* printf("IDs: %d %d.\n", gd->glyphID, gd->newID); */

        if (gd->nbytes != 0) {
            z = GetInt16(gd->ptr, 2);
            if (z < xMin) xMin = z;

            z = GetInt16(gd->ptr, 4);
            if (z < yMin) yMin = z;
        
            z = GetInt16(gd->ptr, 6);
            if (z > xMax) xMax = z;
        
            z = GetInt16(gd->ptr, 8);
            if (z > yMax) yMax = z;
        }

        if (gd->compflag == 0) {                            /* non-composite glyph */
            if (gd->npoints > maxPoints) maxPoints = gd->npoints;
            if (gd->ncontours > maxContours) maxContours = gd->ncontours;
        } else {                                            /* composite glyph */
            if (gd->npoints > maxCompositePoints) maxCompositePoints = gd->npoints;
            if (gd->ncontours > maxCompositeContours) maxCompositeContours = gd->ncontours;
        }
    }

    indexToLocFormat = (glyfLen / 2 > 0xFFFF) ? 1 : 0;
    locaLen = indexToLocFormat ?  (nGlyphs + 1) << 2 : (nGlyphs + 1) << 1;

    // somewhat ...
    TTT_glyf* glyf_2 = new TTT_glyf( glyfLen );
    sal_uInt8* glyfPtr = glyf_2->GetData();

    loca = new TrueTypeTable( T_loca, locaLen );
    sal_uInt8* locaPtr = loca->GetData();

    TTSimpleGlyphMetrics* met = (TTSimpleGlyphMetrics*)scalloc(nGlyphs, sizeof(TTSimpleGlyphMetrics));
    i = 0;

    p1 = glyfPtr;
    p2 = locaPtr;
    for (it = m.begin(); it != m.end(); it++) {
        GlyphData *gd = it->second;
        
        if (gd->compflag) {                       /* re-number all components */
            sal_uInt16 flags, index;
            sal_uInt8 *ptr = gd->ptr + 10;
            do {
                sal_uInt32 j;
                flags = GetUInt16(ptr, 0);
                index = GetUInt16(ptr, 2);
                /* XXX use the sorted array of old to new glyphID mapping and do a binary search */
                for (j = 0; j < nGlyphs; j++) {
                    if (gid[j] == index) {
                        break;
                    }
                }
                /* printf("X: %d -> %d.\n", index, j); */

                PutUInt16((sal_uInt16) j, ptr, 2);

                ptr += 4;

                if (flags & ARG_1_AND_2_ARE_WORDS) {
                    ptr += 4;
                } else {
                    ptr += 2;
                }

                if (flags & WE_HAVE_A_SCALE) {
                    ptr += 2;
                } else if (flags & WE_HAVE_AN_X_AND_Y_SCALE) {
                    ptr += 4;
                } else if (flags & WE_HAVE_A_TWO_BY_TWO) {
                    ptr += 8;
                }
            } while (flags & MORE_COMPONENTS);
        }

        if (gd->nbytes != 0) {
            memcpy(p1, gd->ptr, gd->nbytes);
        }
        if (indexToLocFormat == 1) {
            PutUInt32(p1 - glyfPtr, p2, 0);
            p2 += 4;
        } else {
            PutUInt16((sal_uInt16)((p1 - glyfPtr) >> 1), p2, 0);
            p2 += 2;
        }
        p1 += gd->nbytes;

        /* fill the array of metrics */
        met[i].adv = gd->aw;
        met[i].sb  = gd->lsb;
        i++;
    }

    free(gid);

    if (indexToLocFormat == 1) {
        PutUInt32(p1 - glyfPtr, p2, 0);
    } else {
        PutUInt16((sal_uInt16)((p1 - glyfPtr) >> 1), p2, 0);
    }

    // it is somewhat hard to understand.
    // glyf->rawdata will be freed in GetRawData() before its contents are used.
    // Then it will be overwritten in GetRawData_glyf().
    // glyf->rawdata = glyfPtr;
    // So, discard this glyf_2 here just for now.
    delete glyf_2;

    AddTable(loca);
    
    head = dynamic_cast< TTT_head* >( FindTable( T_head ) ); assert(head != 0);
    sal_uInt8* const pHeadData = head->GetData();
    PutInt16(xMin, pHeadData, 36);
    PutInt16(yMin, pHeadData, 38);
    PutInt16(xMax, pHeadData, 40);
    PutInt16(yMax, pHeadData, 42);
    PutInt16(indexToLocFormat, pHeadData,  50);

    maxp = dynamic_cast< TTT_maxp* >( FindTable( T_maxp ) ); assert(maxp != 0);
    sal_uInt8* const pMaxpData = maxp->GetData();
    PutUInt16((sal_uInt16)nGlyphs, pMaxpData, 4);
    PutUInt16(maxPoints, pMaxpData, 6);
    PutUInt16(maxContours, pMaxpData, 8);
    PutUInt16(maxCompositePoints, pMaxpData, 10);
    PutUInt16(maxCompositeContours, pMaxpData, 12);

#if 0
    /* XXX do not overwrite the existing data. Fix: re-calculate these numbers here */
    PutUInt16(2, maxp->data, 14);                        /* maxZones is always 2       */
    PutUInt16(0, maxp->data, 16);                        /* maxTwilightPoints          */
    PutUInt16(0, maxp->data, 18);                        /* maxStorage                 */
    PutUInt16(0, maxp->data, 20);                        /* maxFunctionDefs            */
    PutUint16(0, maxp->data, 22);                        /* maxInstructionDefs         */
    PutUint16(0, maxp->data, 24);                        /* maxStackElements           */
    PutUint16(0, maxp->data, 26);                        /* maxSizeOfInstructions      */
    PutUint16(0, maxp->data, 28);                        /* maxComponentElements       */
    PutUint16(0, maxp->data, 30);                        /* maxComponentDepth          */
#endif

    /*
     * Generate an htmx table and update hhea table
     */
    hhea = dynamic_cast< TTT_hhea* >( FindTable( T_hhea ) ); assert(hhea != 0);
    hheaPtr = hhea->GetData();
    if (nGlyphs > 2) {
        for (i = nGlyphs - 1; i > 0; i--) {
            if (met[i].adv != met[i-1].adv) break;
        }
        nlsb = nGlyphs - 1 - i;
    }
    hmtxSize = (nGlyphs - nlsb) * 4 + nlsb * 2;

    hmtx = new TrueTypeTable( T_hmtx, hmtxSize );
    hmtxPtr = hmtx->GetData();
    p1 = hmtxPtr;

    for (i = 0; i < nGlyphs; i++) {
        if (i < nGlyphs - nlsb) {
            PutUInt16(met[i].adv, p1, 0);
            PutUInt16(met[i].sb, p1, 2);
            p1 += 4;
        } else {
            PutUInt16(met[i].sb, p1, 0);
            p1 += 2;
        }
    }

    AddTable( hmtx );
    PutUInt16((sal_uInt16)(nGlyphs - nlsb), hheaPtr, 34);

    free(met);
}

void TrueTypeCreator::CreateTables()
{
    ProcessTables();
    /* ProcessTables() adds 'loca' and 'hmtx' */
    
    mpFontDirectory = new TTT_FontDirectory( maTableMap );

    sal_uInt32 checkSumAdjustment = mpFontDirectory->GetCheckSum();

    TTT_TableMap::iterator it;
    for (it = maTableMap.begin(); it != maTableMap.end(); it++) {
        TrueTypeTable* t = it->second;
        checkSumAdjustment += t->GetCheckSum();
    }

    TTT_head* head = dynamic_cast< TTT_head* >( FindTable( T_head ) );
    sal_uInt8* buf = head->GetData();
    PutUInt32(0xB1B0AFBA - checkSumAdjustment, buf, 8);
}

int TrueTypeCreator::StreamToMemory( sal_uInt8 **ptr, sal_uInt32 *length )
{
    CreateTables();

    sal_uInt32 s = CalculateFileSize();
    sal_uInt8* ttf = (sal_uInt8*)smalloc(s);
    sal_uInt8* p = ttf;

    mpFontDirectory->ExportTo( p );
    p += mpFontDirectory->GetSize();

    TTT_TableMap::iterator it;
    for (it = maTableMap.begin(); it != maTableMap.end(); it++) {
        TrueTypeTable* t = it->second;
        t->ExportTo( p );
        p += t->GetContainerSize();
    }

    *ptr = ttf;
    *length = s;
    
    return SF_OK;
}

int TrueTypeCreator::StreamToFile( const char* fname )
{
    int r;
    FILE* fp;

    if (!fname) return SF_BADFILE;

    CreateTables();

    if ( (fp = fopen(fname, "wb")) == NULL )
        return SF_BADFILE;

    r = mpFontDirectory->ExportTo( fp );
    if ( r == SF_OK ) {
        TTT_TableMap::iterator it;
        for (it = maTableMap.begin(); it != maTableMap.end(); it++) {
            TrueTypeTable* t = it->second;
            r = t->ExportTo( fp );
            if ( r != SF_OK )
                break;
        }
    }

    fclose( fp );
    return r;
}

//----------------------------------------------------------------------

/*
 * TrueTypeTable private methods
 */

#define TABLESIZE_head 54
#define TABLESIZE_hhea 36
#define TABLESIZE_maxp 32



/*    Table         data points to
 * --------------------------------------------
 *    'head'        TABLESIZE_head bytes of memory
 *    'hhea'        TABLESIZE_hhea bytes of memory
 *    'loca'        tdata_loca struct
 *    'maxp'        TABLESIZE_maxp bytes of memory
 *    'glyf'        list of GlyphData structs (defined in sft.h)
 *    'name'        list of NameRecord structs (defined in sft.h)
 *    'post'        tdata_post struct
 *
 */



//----------------------------------------------------------------------

/*
 * TrueTypeTable public methods
 */

/* Note: Type42 fonts only need these tables:
 *        head, hhea, loca, maxp, cvt, prep, glyf, hmtx, fpgm
 *
 * Microsoft required tables
 *        cmap, glyf, head, hhea, hmtx, loca, maxp, name, post, OS/2
 *
 * Apple required tables
 *        cmap, glyf, head, hhea, hmtx, loca, maxp, name, post
 *
 */

TrueTypeTable_base::TrueTypeTable_base( sal_uInt32 nSize, const sal_uInt8* pData )
    : mnSize( nSize ),
      mpData( NULL )
{
    if ( nSize ) {
        mpData = static_cast< sal_uInt8* >( calloc( 1, GetContainerSize() ) );
        assert( mpData != 0 );

        if ( pData )
            memcpy( mpData, pData, nSize );
    }
}
    
TrueTypeTable_base::TrueTypeTable_base( sal_uInt32 nSize, const sal_uInt8* pData, sal_uInt32 nExpectedSize )
    : mnSize( nSize ),
      mpData( NULL )
{
    if ( nSize && nSize == nExpectedSize ) {
        mpData = static_cast< sal_uInt8* >( calloc( 1, GetContainerSize() ) );
        assert( mpData != 0 );

        if ( pData )
            memcpy( mpData, pData, nSize );
    }
    else
        nSize = 0;
}

sal_uInt32 TrueTypeTable_base::GetCheckSum()
{
    sal_uInt32 sum = 0;
    sal_uInt32 *ptr = (sal_uInt32 *) mpData;
    sal_uInt32 *endptr = ptr + GetContainerSize() / 4;

    // should we take into account of endian?
    // *ptr or GetUInt32() ?

    while (ptr < endptr)
        sum += *ptr++;

    return sum;
}

int TrueTypeTable_base::ExportTo( sal_uInt8* p )
{
    memcpy( p, mpData, GetContainerSize() );
    return SF_OK;
}

int TrueTypeTable_base::ExportTo( FILE* fp )
{
    sal_uInt32 length = GetContainerSize();

    if ( fwrite( mpData, 1, length, fp ) != length )
        return SF_FILEIO;
    else
        return SF_OK;
}

//----------------------------------------------------------------------

TTT_FontDirectory::TTT_FontDirectory( const TTT_TableMap& aTableMap )
    : TrueTypeTable_base( 12 + 16 * aTableMap.size() )
{
    sal_uInt16 searchRange=1, entrySelector=0, rangeShift;
    sal_uInt16 numTables = (sal_uInt16) aTableMap.size();

    do {
        searchRange *= 2;
        entrySelector++;
    } while (searchRange <= numTables);

    searchRange *= 8;
    entrySelector--;
    rangeShift = numTables * 16 - searchRange;

    /* Offset Subtable */
    sal_uInt8* buf = mpData;
    static const sal_uInt32 T_true = 0x74727565;        /* 'true' */
    PutUInt32(T_true, buf, 0);
    PutUInt16(numTables, buf, 4);
    PutUInt16(searchRange, buf, 6);
    PutUInt16(entrySelector, buf, 8);
    PutUInt16(rangeShift, buf, 10);

    /* Table Directory */
    sal_uInt32 offset = 12 + 16 * numTables;
    int pos = 12;
    TTT_TableMap::const_iterator it;
    for ( it = aTableMap.begin(); it != aTableMap.end(); it++ ) {
        TrueTypeTable* t = it->second;

        PutUInt32(t->GetTag(), buf, pos);
        PutUInt32(t->GetCheckSum(), buf, pos + 4);
        PutUInt32(offset, buf, pos + 8);
        PutUInt32(t->GetSize(), buf, pos + 12);

        offset += t->GetContainerSize();
        pos += 16;
    }
}

//----------------------------------------------------------------------

TrueTypeTable::TrueTypeTable( sal_uInt32 nTag, sal_uInt32 nSize, const sal_uInt8* pData )
    : TrueTypeTable_base( nSize, pData ),
      mnTag( nTag )
{
}
    
TrueTypeTable::TrueTypeTable( sal_uInt32 nTag, sal_uInt32 nSize, const sal_uInt8* pData, sal_uInt32 nExpectedSize )
    : TrueTypeTable_base( nSize, pData, nExpectedSize ),
      mnTag( nTag )
{
}

TrueTypeTable::~TrueTypeTable()
{
    if ( mpData )
        free( mpData );
}

//----------------------------------------------------------------------

TTT_head::TTT_head( sal_uInt32 fontRevision,
                    sal_uInt16 flags,
                    sal_uInt16 unitsPerEm,
                    const sal_uInt8* created,
                    sal_uInt16 macStyle,
                    sal_uInt16 lowestRecPPEM,
                    sal_Int16  fontDirectionHint )
    : TrueTypeTable( T_head, TABLESIZE_head )
{
    assert(created);
    sal_uInt8* ptr = mpData;
    PutUInt32(0x00010000, ptr, 0);             /* version */
    PutUInt32(fontRevision, ptr, 4);
    PutUInt32(0x5F0F3CF5, ptr, 12);            /* magic number */
    PutUInt16(flags, ptr, 16);
    PutUInt16(unitsPerEm, ptr, 18);
    memcpy(ptr+20, created, 8);                   /* Created Long Date */
    memset(ptr+28, 0, 8);                         /* Modified Long Date */
    PutUInt16(macStyle, ptr, 44);
    PutUInt16(lowestRecPPEM, ptr, 46);
    PutUInt16(fontDirectionHint, ptr, 48);
    PutUInt16(0, ptr, 52);                     /* glyph data format: 0 */
}

//----------------------------------------------------------------------

TTT_hhea::TTT_hhea( sal_Int16  ascender,
                    sal_Int16  descender,
                    sal_Int16  linegap,
                    sal_Int16  caretSlopeRise,
                    sal_Int16  caretSlopeRun )
    : TrueTypeTable( T_hhea, TABLESIZE_hhea )
{
    sal_uInt8* ptr = mpData;
    PutUInt32(0x00010000, ptr, 0);             /* version */
    PutUInt16(ascender, ptr, 4);
    PutUInt16(descender, ptr, 6);
    PutUInt16(linegap, ptr, 8);
    PutUInt16(caretSlopeRise, ptr, 18);
    PutUInt16(caretSlopeRun, ptr, 20);
    PutUInt16(0, ptr, 22);                     /* reserved 1 */
    PutUInt16(0, ptr, 24);                     /* reserved 2 */
    PutUInt16(0, ptr, 26);                     /* reserved 3 */
    PutUInt16(0, ptr, 28);                     /* reserved 4 */
    PutUInt16(0, ptr, 30);                     /* reserved 5 */
    PutUInt16(0, ptr, 32);                     /* metricDataFormat */
}

//----------------------------------------------------------------------

TTT_maxp::TTT_maxp( int size, const sal_uInt8* maxp )
    : TrueTypeTable( T_maxp, size, maxp, TABLESIZE_maxp )
{
}

//----------------------------------------------------------------------

TTT_glyf::TTT_glyf()
    : TrueTypeTable( T_glyf ),
      mnLastID(0)
{
}

TTT_glyf::TTT_glyf( sal_uInt32 nSize )
    : TrueTypeTable( T_glyf, nSize ),
      mnLastID(0)
{
}

TTT_glyf::~TTT_glyf()
{
    GlyphDataPtrMap::const_iterator it;
    for (it = maGlyphDataPtrMap.begin(); it != maGlyphDataPtrMap.end(); it++) {
        GlyphData* p = it->second;
        if (p)
            free( p );
    }

    if ( mpData )
        free( mpData );
}

sal_uInt32 TTT_glyf::glyfAdd( GlyphData *glyphdata, TrueTypeFont *fnt )
{
    int ret, ncomponents;

    if (!glyphdata) return (sal_uInt32)~0;

    std::vector< sal_uInt32 > glyphlist;
    ncomponents = GetTTGlyphComponents( fnt, glyphdata->glyphID, glyphlist );

    ret = mnLastID;
    glyphdata->newID = mnLastID++;
    maGlyphDataPtrMap[ glyphdata->glyphID ] = glyphdata;

    if (ncomponents > 1 && glyphlist.size() > 1 )
    {
        std::vector< sal_uInt32 >::const_iterator it = glyphlist.begin();
        ++it;
        /* glyphData->glyphID is always the first glyph on the list */
        for (; it !=  glyphlist.end(); it++) {
            sal_uInt32 currentID = *it;
            GlyphData *gd;

            if (maGlyphDataPtrMap.find( currentID ) == maGlyphDataPtrMap.end()) {
                gd = GetTTRawGlyphData( fnt, currentID );
                gd->newID = mnLastID++;
                maGlyphDataPtrMap[ currentID ] = gd;
            }
        }
    }

    return ret;
}

sal_uInt32 TTT_glyf::GetSize()
{
    if ( 0 < mnSize )
        return mnSize;

    GlyphDataPtrMap::const_iterator it;
    sal_uInt32 nbytes = 0;

    for (it = maGlyphDataPtrMap.begin(); it != maGlyphDataPtrMap.end(); it++) {
        nbytes += it->second->nbytes;
    }        

    mnSize = nbytes;
    return mnSize;
}

int TTT_glyf::ExportTo( sal_uInt8* p )
{
    GlyphDataPtrMap::const_iterator it;

    for (it = maGlyphDataPtrMap.begin(); it != maGlyphDataPtrMap.end(); it++) {
        GlyphData* g = it->second;
        sal_uInt32 n = g->nbytes;
        if ( n != 0 ) {
            memcpy( p, g->ptr, n );
            p += n;
        }
    }

    return SF_OK;
}

int TTT_glyf::ExportTo( FILE* fp )
{
    GlyphDataPtrMap::const_iterator it;

    for (it = maGlyphDataPtrMap.begin(); it != maGlyphDataPtrMap.end(); it++) {
        GlyphData* g = it->second;
        sal_uInt32 n = g->nbytes;
        if ( n != 0 )
            if ( fwrite( g->ptr, 1, n, fp ) != n )
                return SF_FILEIO;
    }

    return SF_OK;
}

//----------------------------------------------------------------------

sal_uInt32 CmapSubTableExporter0::GetContainerSize()
{
    return 262;
}

int CmapSubTableExporter0::ExportTo( sal_uInt8* p )
{
    sal_uInt32 n = mpCmapEntryVector->size();
    sal_uInt32 i, j;
    sal_uInt16 g;

    PutUInt16( 0, p, 0 );
    PutUInt16( 262, p, 2 );
    PutUInt16( 0, p, 4 );
    p += 6;

    for ( i = 0; i < 256; i++ ) {
        g = 0;
        for ( j = 0; j < n; j++ ) {
            CmapEntry& entry = (*mpCmapEntryVector)[j];
            if ( entry.mnChar == i ) {
                g = (sal_uInt16) entry.mnGlyph;;
            }
        }
        *p++ = (sal_uInt8) g;
    }

    return SF_OK;
}

int CmapSubTableExporter0::ExportTo( FILE* fp )
{
    int ret = SF_OK;
    sal_uInt32 size = GetContainerSize();
    sal_uInt8* buf = (sal_uInt8*) smalloc( size );
    ExportTo( buf );

    if ( fwrite( buf, 1, size, fp ) != size )
        ret = SF_FILEIO;

    free(buf);
    return ret;
}

//----------------------------------------------------------------------

sal_uInt32 CmapSubTableExporter6::GetContainerSize()
{
    return 10 + mpCmapEntryVector->size() * 2;
}

int CmapSubTableExporter6::ExportTo( sal_uInt8* p )
{
    sal_uInt32 n = mpCmapEntryVector->size();
    sal_uInt32 i, j;
    sal_uInt16 g;

    PutUInt16( 6, p, 0 );
    PutUInt16( (sal_uInt16)(n*2+10), p, 2 );
    PutUInt16( 0, p, 4 );
    PutUInt16( 0, p, 6 );
    PutUInt16( (sal_uInt16)(n), p, 8 );
    p += 10;

    for ( i = 0; i < n; i++ ) {
        g = 0;
        for ( j = 0; j < n; j++ ) {
            CmapEntry& entry = (*mpCmapEntryVector)[j];
            if ( entry.mnChar == i ) {
                g = (sal_uInt16) entry.mnGlyph;;
            }
        }
        PutUInt16( g, p, 0 );
        p += 2;
    }

    return SF_OK;
}

int CmapSubTableExporter6::ExportTo( FILE* fp )
{
    int ret = SF_OK;
    sal_uInt32 size = GetContainerSize();
    sal_uInt8* buf = (sal_uInt8*) smalloc( size );
    ExportTo( buf );

    if ( fwrite( buf, 1, size, fp ) != size )
        ret = SF_FILEIO;

    free(buf);
    return ret;
}

//----------------------------------------------------------------------

#define CMAP_ENTRY_INDEX_UPPER_BITS 8
#define CMAP_ENTRY_INDEX_LOWER_BITS 8

CmapSubTable::CmapSubTable()
    : maCmapEntryVector( CMAP_ENTRY_INDEX_UPPER_BITS, CMAP_ENTRY_INDEX_LOWER_BITS ),
      mnSize( 0 )
{
}

void CmapSubTable::AddEntry( sal_uInt32 nChar, sal_uInt32 nGlyph )
{
    CmapEntry entry = { nChar, nGlyph };
    maCmapEntryVector.push_back( entry );
}

void CmapSubTable::DetermineFormatType()
{
    // is it better to find the maximum glyph number, instead of using the last one?

    if ( ! moExporter.getCmapEntryVector() ) {
        CmapEntry& entry = maCmapEntryVector.back();

        if ( entry.mnGlyph <= 0xff )
            moExporter = CmapSubTableExporter0( &maCmapEntryVector );  // format 0
        else
            moExporter = CmapSubTableExporter6( &maCmapEntryVector );  // format 6
    }
}

sal_uInt32 CmapSubTable::GetContainerSize()
{
    if ( 0 == mnSize ) {
        DetermineFormatType();
        mnSize = moExporter.GetContainerSize();
    }

    return mnSize;
}

int CmapSubTable::ExportTo( sal_uInt8* p )
{
    DetermineFormatType();
    return moExporter.ExportTo( p );
}

int CmapSubTable::ExportTo( FILE* fp )
{
    DetermineFormatType();
    return moExporter.ExportTo( fp );
}


//----------------------------------------------------------------------

TTT_cmap::TTT_cmap()
    : TrueTypeTable( T_cmap )
{
}

TTT_cmap::~TTT_cmap()
{

}

void TTT_cmap::cmapAdd( sal_uInt32 id, sal_uInt32 c, sal_uInt32 g )
{
    CmapSubTable& t = maCmapSubTableMap[ id ];
    t.AddEntry( c, g );

//    if (i < t->n) {
//        memmove(s+i+1, s+i, (t->n-i) * sizeof(s[0]));  // ???
//        memmove(s+i+1, s+i, t->n-i);
//    }
}

sal_uInt32 TTT_cmap::GetSize()
{
    if ( 0 < mnSize )
        return mnSize;

    sal_uInt32 tlen = 0;
    CmapSubTableMap::iterator it;

    for ( it = maCmapSubTableMap.begin(); it != maCmapSubTableMap.end(); it++ ) {
        CmapSubTable& t = it->second;
        tlen += t.GetContainerSize();
    }

    mnSize = 4 + maCmapSubTableMap.size() * 8 + tlen;
    return mnSize;
}

int TTT_cmap::ExportTo( sal_uInt8* p )
{
    sal_uInt8* cmap = p;
    sal_uInt16 n = (sal_uInt16) maCmapSubTableMap.size();

    PutUInt16( 0, cmap, 0 );
    PutUInt16( n, cmap, 2 );
    p += 4;

    sal_uInt32 coffset = 4 + n * 8;
    CmapSubTableMap::iterator it;

    for ( it = maCmapSubTableMap.begin(); it != maCmapSubTableMap.end(); it++ ) {
        sal_uInt32 id = it->first;
        CmapSubTable& t = it->second;

        PutUInt16( (sal_uInt16)(id >> 16), p, 0 );
        PutUInt16( (sal_uInt16)(id & 0xFF), p, 2 );
        PutUInt32( coffset, p, 4 );

        t.ExportTo( cmap + coffset );
        coffset += t.GetContainerSize();

        p += 8;
    }

    return SF_OK;
}

int TTT_cmap::ExportTo( FILE* fp )
{
    sal_uInt16 n = (sal_uInt16) maCmapSubTableMap.size();
    sal_uInt8 buf[8];

    PutUInt16( 0, buf, 0 );
    PutUInt16( n, buf, 2 );

    if ( fwrite( buf, 1, 4, fp ) != 4 )
        return SF_FILEIO;

    sal_uInt32 coffset = 4 + n * 8;
    CmapSubTableMap::iterator it;

    for (it = maCmapSubTableMap.begin(); it != maCmapSubTableMap.end(); it++) {
        sal_uInt32 id = it->first;
        CmapSubTable& t = it->second;

        PutUInt16( (sal_uInt16)(id >> 16), buf, 0 );
        PutUInt16( (sal_uInt16)(id & 0xFF), buf, 2 );
        PutUInt32( coffset, buf, 4 );

        if ( fwrite( buf, 1, 8, fp ) != 8 )
            return SF_FILEIO;

        coffset += t.GetContainerSize();
    }

    for ( it = maCmapSubTableMap.begin(); it != maCmapSubTableMap.end(); it++ ) {
        int ret;
        CmapSubTable& t = it->second;

        ret = t.ExportTo( fp );
        if ( ret != SF_OK )
            return ret;

        coffset += t.GetContainerSize();
    }

    return SF_OK;
}

//----------------------------------------------------------------------

TTT_name::TTT_name( int n, NameRecord *nr )
    : TrueTypeTable( T_name ),
      mpNameRecords( NULL )
{
    while ( 0 < n-- ) {
        maNameRecordPtrMap[ NameRecordKey( nr ) ] = NameRecordValue( nr );
        nr++;
    }
}

TTT_name::TTT_name( TrueTypeFont* ttf )
    : TrueTypeTable( T_name ),
      mpNameRecords( NULL )
{
    int n = GetTTNameRecords( ttf, &mpNameRecords );
    NameRecord *nr = mpNameRecords;

    while ( 0 < n-- ) {
        maNameRecordPtrMap[ NameRecordKey( nr ) ] = NameRecordValue( nr );
        nr++;
    }
}

TTT_name::~TTT_name()
{
    // something 
    int n = maNameRecordPtrMap.size();

    if ( mpNameRecords )
        DisposeNameRecords( mpNameRecords, n );

    // some destractors here

    NameRecordPtrMap::iterator it;
    for (it = maNameRecordPtrMap.begin(); it != maNameRecordPtrMap.end(); it++) {
        NameRecordValue& v = it->second;

        // some free() here

    }
}

void TTT_name::nameAdd( NameRecord *nr )
{
    maNameRecordPtrMap[ NameRecordKey( nr ) ] = NameRecordValue( nr );
}

sal_uInt32 TTT_name::GetSize()
{
    if ( 0 < mnSize )
        return mnSize;

    sal_Int16 n = (sal_Int16) maNameRecordPtrMap.size();
    int stringLen = 0;

    NameRecordPtrMap::iterator it;
    for (it = maNameRecordPtrMap.begin(); it != maNameRecordPtrMap.end(); it++) {
        NameRecordValue& v = it->second;
        stringLen += v.slen;
    }

//    if (stringLen > 65535) {
//        return TTCR_NAMETOOLONG;
//    }

    mnSize = stringLen + 12 * n + 6;
    return mnSize;
}

int TTT_name::ExportTo( sal_uInt8* p )
{
    sal_Int16 n = (sal_Int16) maNameRecordPtrMap.size();
    sal_uInt8 *p1, *p2;
    sal_uInt16 offset;

    PutUInt16( 0, p, 0 );
    PutUInt16( n, p, 2 );
    PutUInt16( (sal_uInt16)(6 + 12 * n), p, 4 );

    p1 = p + 6;
    p2 = p1 + 12 * n;
    offset = 0;

    NameRecordPtrMap::iterator it;
    for (it = maNameRecordPtrMap.begin(); it != maNameRecordPtrMap.end(); it++) {
        NameRecordKey   k = it->first;
        NameRecordValue v = it->second;

        PutUInt16( k.platformID, p1, 0 );
        PutUInt16( k.encodingID, p1, 2 );
        PutUInt16( k.languageID, p1, 4 );
        PutUInt16( k.nameID,     p1, 6 );
        PutUInt16( v.slen,       p1, 8 );
        PutUInt16( offset,         p1, 10 );
        //PutUInt16( (sal_uInt16)(p2 - (name + 6 + 12 * n)), p1, 10 );
        memcpy( p2, v.sptr, v.slen );
        /* {int j; for(j=0; j<nr[i].slen; j++) printf("%c", nr[i].sptr[j]); printf("\n"); }; */

        offset += v.slen;
        p2 += v.slen;
        p1 += 12;
    }

    return SF_OK;
}

int TTT_name::ExportTo( FILE* fp )
{
    sal_Int16 n = (sal_Int16) maNameRecordPtrMap.size();
    sal_uInt8 buf[12];
    sal_uInt16 offset;

    PutUInt16( 0, buf, 0 );
    PutUInt16( n, buf, 2 );
    PutUInt16( (sal_uInt16)(6 + 12 * n), buf, 4 );

    if ( fwrite( buf, 1, 6, fp ) != 6 )
        return SF_FILEIO;

    offset = 0;

    NameRecordPtrMap::iterator it;
    for (it = maNameRecordPtrMap.begin(); it != maNameRecordPtrMap.end(); it++) {
        NameRecordKey   k = it->first;
        NameRecordValue v = it->second;

        PutUInt16( k.platformID, buf, 0 );
        PutUInt16( k.encodingID, buf, 2 );
        PutUInt16( k.languageID, buf, 4 );
        PutUInt16( k.nameID,     buf, 6 );
        PutUInt16( v.slen,       buf, 8 );
        PutUInt16( offset,         buf, 10 );

        if ( fwrite( buf, 1, 12, fp ) != 12 )
            return SF_FILEIO;

        offset += v.slen;
    }

    for (it = maNameRecordPtrMap.begin(); it != maNameRecordPtrMap.end(); it++) {
        NameRecordValue& v = it->second;

        if ( fwrite( v.sptr, 1, v.slen, fp ) != v.slen )
            return SF_FILEIO;
    }

    return SF_OK;
}

//----------------------------------------------------------------------

TTT_post::TTT_post(sal_uInt32 format,
                   sal_uInt32 italicAngle,
                   sal_Int16 underlinePosition,
                   sal_Int16 underlineThickness,
                   sal_uInt32 isFixedPitch)

    : TrueTypeTable( T_post )

    // some additional initializers here

{
    format;
    italicAngle;
    underlinePosition;
    underlineThickness;
    isFixedPitch;
}

int TTT_post::ExportTo( sal_uInt8* p )
{
    // some implements here

    p;

    return SF_OK;
}

int TTT_post::ExportTo( FILE* fp )
{
    // some implements here

    fp;

    return SF_OK;
}

} // namespace vcl
               
#ifdef TEST_TTCR
int main(void)
{
    TrueTypeCreator ttcr;
    sal_uInt8 *t1, *t2, *t3, *t4, *t5, *t6, *t7;

    t1 = malloc(1000); memset(t1, 'a', 1000);
    t2 = malloc(2000); memset(t2, 'b', 2000);
    t3 = malloc(3000); memset(t3, 'c', 3000);
    t4 = malloc(4000); memset(t4, 'd', 4000);
    t5 = malloc(5000); memset(t5, 'e', 5000);
    t6 = malloc(6000); memset(t6, 'f', 6000);
    t7 = malloc(7000); memset(t7, 'g', 7000);

    ttcr.AddTable( new TrueTypeTable(0x6D617870, 1000, t1) );
    ttcr.AddTable( new TrueTypeTable(0x4F532F32, 2000, t2) );
    ttcr.AddTable( new TrueTypeTable(0x636D6170, 3000, t3) );
    ttcr.AddTable( new TrueTypeTable(0x6C6F6361, 4000, t4) );
    ttcr.AddTable( new TrueTypeTable(0x68686561, 5000, t5) );
    ttcr.AddTable( new TrueTypeTable(0x676C7966, 6000, t6) );
    ttcr.AddTable( new TrueTypeTable(0x6B65726E, 7000, t7) );

    free(t1);
    free(t2);
    free(t3);
    free(t4);
    free(t5);
    free(t6);
    free(t7);

    ttcr.StreamToFile("ttcrout.ttf");

    return 0;
}
#endif
