/*************************************************************************
 *
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 * 
 * Copyright 2000, 2010 Oracle and/or its affiliates.
 *
 * OpenOffice.org - a multi-platform office productivity suite
 *
 * This file is part of OpenOffice.org.
 *
 * OpenOffice.org is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 3
 * only, as published by the Free Software Foundation.
 *
 * OpenOffice.org is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3 for more details
 * (a copy is included in the LICENSE file that accompanied this code).
 *
 * You should have received a copy of the GNU Lesser General Public License
 * version 3 along with OpenOffice.org.  If not, see
 * <http://www.openoffice.org/license.html>
 * for a copy of the LGPLv3 License.
 *
 ************************************************************************/

/**
 *
 * @file ttcr.h
 * @brief TrueType font creator
 * @author Alexander Gelfenbain
 */

#ifndef __TTCR_H
#define __TTCR_H

#include <osl/endian.h>
#include <map>

#include "sft.hxx"
#include "vec.hxx"

/* standard TrueType table tags */
#define T_CFF   0x43464620
#define T_OS2   0x4F532F32
#define T_cmap  0x636D6170
#define T_cvt   0x63767420
#define T_fpgm  0x6670676D
#define T_glyf  0x676C7966
#define T_gsub  0x47535542
#define T_head  0x68656164
#define T_hhea  0x68686561
#define T_hmtx  0x686D7478
#define T_kern  0x6B65726E
#define T_loca  0x6C6F6361
#define T_maxp  0x6D617870
#define T_name  0x6E616D65
#define T_post  0x706F7374
#define T_prep  0x70726570
#define T_vhea  0x76686561
#define T_vmtx  0x766D7478

namespace vcl
{

inline void *smalloc(sal_uInt32 size)
{
    void *res = malloc(size);
    assert(res != 0);
    return res;
}

inline void *scalloc(sal_uInt32 n, sal_uInt32 size)
{
    void *res = calloc(n, size);
    assert(res != 0);
    return res;
}

// The data types used in TrueType font files are aligned in big-endian
inline sal_Int16 GetInt16( const sal_uInt8* ptr, sal_uInt8 offset)
{
    assert(ptr);
    ptr += offset;
#ifdef OSL_BIGENDIAN
    return *(static_cast<sal_Int16*> ptr);
#else
    return ( ptr[0] << 8 | ptr[1] );
#endif
}

inline sal_uInt16 GetUInt16( const sal_uInt8* ptr, sal_uInt8 offset)
{
    assert(ptr);
    ptr += offset;
#ifdef OSL_BIGENDIAN
    return *(static_cast<sal_uInt16*> ptr);
#else
    return ( ptr[0] << 8 | ptr[1] );
#endif
}

inline sal_Int32 GetInt32( const sal_uInt8* ptr, sal_uInt8 offset)
{
    assert(ptr);
    ptr += offset;
#ifdef OSL_BIGENDIAN
    return *(static_cast<sal_Int32*> ptr);
#else
    return ( ( ( ptr[0] << 8 | ptr[1] ) << 8 | ptr[2] ) << 8 | ptr[3] );
#endif
}

inline sal_uInt32 GetUInt32( const sal_uInt8* ptr, sal_uInt8 offset)
{
    assert(ptr);
    ptr += offset;
#ifdef OSL_BIGENDIAN
    return *(static_cast<sal_uInt32*> ptr);
#else
    return ( ( ( ptr[0] << 8 | ptr[1] ) << 8 | ptr[2] ) << 8 | ptr[3] );
#endif
}

inline void PutInt16(sal_Int16 val, sal_uInt8 *ptr, sal_uInt8 offset)
{
    assert(ptr);
    ptr += offset;
#ifdef OSL_BIGENDIAN
    *(static_cast<sal_Int16*> ptr) = val;
#else
    ptr[1] = (sal_uInt8)(val & 0xFF);
    ptr[0] = (sal_uInt8)((val >> 8) & 0xFF);
#endif
}
 
inline void PutUInt16(sal_uInt16 val, sal_uInt8 *ptr, sal_uInt8 offset)
{
    assert(ptr);
    ptr += offset;
#ifdef OSL_BIGENDIAN
    *(static_cast<sal_uInt16*> ptr) = val;
#else
    ptr[1] = (sal_uInt8)(val & 0xFF);
    ptr[0] = (sal_uInt8)((val >> 8) & 0xFF);
#endif
}

inline void PutUInt32(sal_uInt32 val, sal_uInt8 *ptr, sal_uInt8 offset)
{
    assert(ptr);
    ptr += offset;
#ifdef OSL_BIGENDIAN
    *(static_cast<sal_Int32*> ptr) = val;
#else
    ptr[3] = (sal_uInt8)(val & 0xFF);
    ptr[2] = (sal_uInt8)((val >>= 8) & 0xFF);
    ptr[1] = (sal_uInt8)((val >>= 8) & 0xFF);
    ptr[0] = (sal_uInt8)((val >>= 8) & 0xFF);
#endif
}

inline void PutInt32(sal_Int32 val, sal_uInt8 *ptr, sal_uInt8 offset)
{
    assert(ptr);
    ptr += offset;
#ifdef OSL_BIGENDIAN
    *(static_cast<sal_uInt32*> ptr) = val;
#else
    ptr[3] = (sal_uInt8)(val & 0xFF);
    ptr[2] = (sal_uInt8)((val >>= 8) & 0xFF);
    ptr[1] = (sal_uInt8)((val >>= 8) & 0xFF);
    ptr[0] = (sal_uInt8)((val >>= 8) & 0xFF);
#endif
}

//----------------------------------------------------------------------

class TrueTypeTable;

typedef std::map< const sal_uInt32, TrueTypeTable* > TTT_TableMap;

//----------------------------------------------------------------------

class TrueTypeTable_base
{
protected:
    sal_uInt32     mnSize;
    sal_uInt8*     mpData;

public:
    TrueTypeTable_base( sal_uInt32 nSize = 0, const sal_uInt8* pData = NULL );
    TrueTypeTable_base( sal_uInt32 nSize, const sal_uInt8* pData, sal_uInt32 nExpectedSize);

    sal_uInt32 GetCheckSum();
    int ExportTo( sal_uInt8* p );
    int ExportTo( FILE* fp );

    sal_uInt32 GetSize() { return mnSize; }
    sal_uInt8* GetData() const { return mpData; }
    sal_uInt32 GetContainerSize() { return (GetSize() + 3) & (sal_uInt32) ~3; }
};

//----------------------------------------------------------------------

class TTT_FontDirectory : public TrueTypeTable_base
{
public:
    TTT_FontDirectory( const TTT_TableMap& aTableMap );
};

//----------------------------------------------------------------------

class TrueTypeTable : public TrueTypeTable_base
{
protected:
    sal_uInt32     mnTag;

public:
    TrueTypeTable( sal_uInt32 nTag, sal_uInt32 nSize = 0, const sal_uInt8* pData = NULL );
    TrueTypeTable( sal_uInt32 nTag, sal_uInt32 nSize, const sal_uInt8* pData, sal_uInt32 nExpectedSize);
    virtual ~TrueTypeTable();

    sal_uInt32 GetTag()  const { return mnTag; }
};

//----------------------------------------------------------------------

class TTT_head : public TrueTypeTable
{
public:
    TTT_head( sal_uInt32 fontRevision,
              sal_uInt16 flags,
              sal_uInt16 unitsPerEm,
              const sal_uInt8* created,
              sal_uInt16 macStyle,
              sal_uInt16 lowestRecPPEM,
              sal_Int16  fontDirectionHint );
};

//----------------------------------------------------------------------

class TTT_hhea : public TrueTypeTable
{
public:
    TTT_hhea( sal_Int16  ascender,
              sal_Int16  descender,
              sal_Int16  linegap,
              sal_Int16  caretSlopeRise,
              sal_Int16  caretSlopeRun );
};

//----------------------------------------------------------------------

class TTT_maxp : public TrueTypeTable
{
public:
    TTT_maxp( int size, const sal_uInt8* maxp );
};

//----------------------------------------------------------------------

typedef std::map< sal_uInt32, GlyphData* > GlyphDataPtrMap;

class TTT_glyf : public TrueTypeTable
{
    GlyphDataPtrMap   maGlyphDataPtrMap;
    sal_uInt32        mnLastID;

public:
    TTT_glyf();
    TTT_glyf( sal_uInt32 nSize );
    virtual ~TTT_glyf();

    sal_uInt32 glyfAdd( GlyphData *glyphdata, TrueTypeFont *fnt );
    sal_uInt32 GetSize();

    int ExportTo( sal_uInt8* p );
    int ExportTo( FILE* fp );

    const GlyphDataPtrMap& GetGlyphDataPtrMap() const { return maGlyphDataPtrMap; }
};

//----------------------------------------------------------------------

typedef struct
{
    sal_uInt32  mnChar;
    sal_uInt32  mnGlyph;
} CmapEntry;

typedef Vec< CmapEntry > CmapEntryVector;

class CmapSubTableExporter
{
protected:
    CmapEntryVector*       mpCmapEntryVector;

public:
    CmapSubTableExporter()
        : mpCmapEntryVector( NULL ) {}

    CmapSubTableExporter( CmapEntryVector* pCmapEntryVector )
        : mpCmapEntryVector( pCmapEntryVector ) {}

    virtual sal_uInt32 GetContainerSize() { assert( 0 ); return 0; }
    virtual int ExportTo( sal_uInt8* p ) { assert( 0 ); p; return SF_BADARG; }
    virtual int ExportTo( FILE* fp ) { assert( 0 ); fp; return SF_BADARG; }

    CmapEntryVector* getCmapEntryVector()
        { return mpCmapEntryVector; }
};

//----------------------------------------------------------------------

class CmapSubTableExporter0 : public CmapSubTableExporter
{
public:
    CmapSubTableExporter0( CmapEntryVector* pCmapEntryVector )
        : CmapSubTableExporter( pCmapEntryVector ) {}

    virtual sal_uInt32 GetContainerSize();
    virtual int ExportTo( sal_uInt8* p );
    virtual int ExportTo( FILE* fp );
};

//----------------------------------------------------------------------

class CmapSubTableExporter6 : public CmapSubTableExporter
{
public:
    CmapSubTableExporter6( CmapEntryVector* pCmapEntryVector )
        : CmapSubTableExporter( pCmapEntryVector ) {}

    virtual sal_uInt32 GetContainerSize();
    virtual int ExportTo( sal_uInt8* p );
    virtual int ExportTo( FILE* fp );
};

//----------------------------------------------------------------------

class CmapSubTable
{
    CmapEntryVector        maCmapEntryVector;
    sal_uInt32             mnSize;
    CmapSubTableExporter   moExporter;

public:
    CmapSubTable();

    void AddEntry( sal_uInt32 nChar, sal_uInt32 nGlyph );
    void DetermineFormatType();
    sal_uInt32 GetContainerSize();
    int ExportTo( sal_uInt8* p );
    int ExportTo( FILE* fp );
};

//----------------------------------------------------------------------

typedef std::map< sal_uInt32, CmapSubTable > CmapSubTableMap;

class TTT_cmap : public TrueTypeTable
{
    CmapSubTableMap      maCmapSubTableMap;

public:
    TTT_cmap();
    virtual ~TTT_cmap();

    void cmapAdd( sal_uInt32 id, sal_uInt32 c, sal_uInt32 g );
    int ExportTo( sal_uInt8* p );
    int ExportTo( FILE* fp );

    sal_uInt32 GetSize();
};

//----------------------------------------------------------------------

class NameRecordKey
{
public:
    sal_uInt16 platformID;  /**< Platform ID                                            */
    sal_uInt16 encodingID;  /**< Platform-specific encoding ID                          */
    sal_uInt16 languageID;  /**< Language ID                                            */
    sal_uInt16 nameID;      /**< Name ID                                                */

    NameRecordKey( NameRecord *nr )
        : platformID( nr->platformID ),
          encodingID( nr->encodingID ),
          languageID( nr->languageID ),
          nameID( nr->nameID ) {}
};

class NameRecordValue
{
public:
    sal_uInt16 slen;        /**< String length in bytes                                 */
    sal_uInt8  *sptr;       /**< Pointer to string data (not zero-terminated!)          */

    NameRecordValue()
        : slen( 0 ),
          sptr( NULL ) {}

    NameRecordValue( NameRecord *nr )
        : slen( nr->slen ),
          sptr( nr->sptr ) {}
};

inline bool operator<( const NameRecordKey& x, const NameRecordKey& y )
{
    if ( x.platformID != y.platformID ) {
        return  x.platformID -  y.platformID;
    } else if ( x.encodingID != y.encodingID ) {
        return  x.encodingID -  y.encodingID;
    } else if ( x.languageID != y.languageID ) {
        return  x.languageID -  y.languageID;
    } else if ( x.nameID != y.nameID ) {
        return x.nameID - y.nameID;
    }
    return 0;
}

typedef std::map< NameRecordKey, NameRecordValue > NameRecordPtrMap;

class TTT_name : public TrueTypeTable
{
    NameRecordPtrMap    maNameRecordPtrMap;
    NameRecord*         mpNameRecords;

public:
    TTT_name( int n, NameRecord *nr );
    TTT_name( TrueTypeFont* ttf );
    virtual ~TTT_name();

    void nameAdd( NameRecord *nr );
    int ExportTo( sal_uInt8* p );
    int ExportTo( FILE* fp );

    sal_uInt32 GetSize(); 
};

//----------------------------------------------------------------------

class TTT_post : public TrueTypeTable
{
    // some members here

public:
    TTT_post(sal_uInt32 format,
             sal_uInt32 italicAngle,
             sal_Int16 underlinePosition,
             sal_Int16 underlineThickness,
             sal_uInt32 isFixedPitch);

    int ExportTo( sal_uInt8* p );
    int ExportTo( FILE* fp );
};

//----------------------------------------------------------------------

class TrueTypeCreator
{
    void ProcessTables();
    void CreateTables();

protected:
    TTT_TableMap        maTableMap;
    TTT_FontDirectory*  mpFontDirectory;

public:
    TrueTypeCreator();
    ~TrueTypeCreator();

    void AddTable( TrueTypeTable* pTable );
    void RemoveTable( sal_uInt32 nTag );
    TrueTypeTable* FindTable( sal_uInt32 nTag );
    sal_uInt32 CalculateFileSize();

    int StreamToMemory( sal_uInt8 **ptr, sal_uInt32 *length );
    int StreamToFile( const char* fname );
};

} // namespace                  

#endif /* __TTCR_H */
