/*************************************************************************
 *
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 * 
 * Copyright 2000, 2010 Oracle and/or its affiliates.
 *
 * OpenOffice.org - a multi-platform office productivity suite
 *
 * This file is part of OpenOffice.org.
 *
 * OpenOffice.org is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License version 3
 * only, as published by the Free Software Foundation.
 *
 * OpenOffice.org is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3 for more details
 * (a copy is included in the LICENSE file that accompanied this code).
 *
 * You should have received a copy of the GNU Lesser General Public License
 * version 3 along with OpenOffice.org.  If not, see
 * <http://www.openoffice.org/license.html>
 * for a copy of the LGPLv3 License.
 *
 ************************************************************************/

#ifndef __VEC_H
#define __VEC_H

#include <assert.h>
#include <string.h>

template< class T >
class Vec;

template< class T >
class Vec_iterator
{
public:
    typedef Vec< T >           container;
    typedef Vec_iterator< T >  iterator;
    typedef unsigned int       size_type;
    typedef T&                 reference;

protected:
    Vec< T >*  mpVec;
    size_type  mnCurrent;

public:
    Vec_iterator()
        : mpVec( NULL ), mnCurrent( 0 ) {}

    Vec_iterator( const iterator& r )
        : mpVec( r.mpVec ), mnCurrent( r.mnCurrent ) {}

    Vec_iterator( container* pVec, size_type  nCurrent )
        : mpVec( pVec ), mnCurrent( nCurrent ) {}

    reference operator*() const {
        return mpVec->operator[]( mnCurrent );
    }

    bool operator==( const Vec_iterator& r ) const { 
        return mnCurrent == r.mnCurrent;
    }

    bool operator!=( const Vec_iterator& r ) const { 
        return mnCurrent != r.mnCurrent;
    }

    iterator& operator++() { 
        mnCurrent++;
        return *this;
    }

    iterator operator++(int) { 
        iterator tmp = *this;
        mnCurrent++;
        return tmp;
    }

};

template< class T >
class Vec {
public:
    typedef T                  value_type;
    typedef T*                 pointer;
    typedef T&                 reference;
    typedef const T&           const_reference;
    typedef unsigned int       size_type;
    typedef Vec_iterator< T >  iterator;

protected:
    size_type  mnUpperSize;
    size_type  mnLowerSize;
    size_type  mnLowerMask;
    size_type  mnLowerBits;
    size_type  mnCount;
    pointer*   mppUpperTable;

    size_type  getUpperValue( size_type index ) {
        size_type n = index >> mnLowerBits;
        assert( 0 <= n && n < mnUpperSize );
        //printf("getUpperValue( %d ) = %d\n", index, n);
        return n;
    }

    size_type  getLowerValue( size_type index ) {
        size_type n = index & mnLowerMask;
        //printf("getLowerValue( %d ) = %d\n", index, n);
        return n;
    }

public:
    Vec( size_type nUpperBits = 8, size_type nLowerBits  = 8 )
        : mnUpperSize( 1 << nUpperBits ),
          mnLowerSize( 1 << nLowerBits ),
          mnLowerMask( mnLowerSize - 1 ),
          mnLowerBits( nLowerBits ),
          mnCount( 0 ) {
        mppUpperTable = new pointer[ mnUpperSize ];
        memset( mppUpperTable, 0, sizeof( pointer ) * mnUpperSize );
    }

    ~Vec() {
        size_type n = mnUpperSize;
        pointer*  p = mppUpperTable;
        //printf("n = %d\n", n);
        while ( 0 < n-- ) {
            //printf("n = %d, %08x\n", n, *p);
            delete[] *p;
            p++;
        }
        delete[] mppUpperTable;
    }

    reference operator[]( size_type index ) {
        size_type up = getUpperValue( index );
        pointer p = mppUpperTable[ up ];
        if ( !p ) {
            p = new value_type[ mnLowerSize ];
            mppUpperTable[ up ] = p;
        }
        size_type lo = getLowerValue( index );
        return p[ lo ];
    }

    void push_back( const_reference value ) {
        reference r = operator[]( mnCount );
        r = value;
        mnCount++;
    }

    bool empty() const
        { return mnCount == 0; }

    size_type size() const
        { return mnCount; }

    iterator begin()
        { return iterator( this, 0 ); }

    iterator end()
        { return iterator( this, mnCount ); }

    reference back() {
        assert( 0 < mnCount );
        return operator[]( mnCount - 1);
    }

    void exportTo( value_type* p ) {
        size_type up = getUpperValue( mnCount );
        size_type lo = getLowerValue( mnCount );
        pointer*  q  = mppUpperTable;

        // This part is for you. Enjoy!

            // repeat this ( up ) times
                // do memcpy( p, *q, mnLowerSize )
                // proceed p and q

            // do memcpy( p, *q, the_rest )
    }
}; 

#endif /* __VEC_H */

